@@ -1,54 +1,62 @@
 from __future__ import unicode_literals
 
-from django import forms, VERSION as DJANGO_VERSION
+from django import forms
+from django.forms.utils import flatatt
 from django.contrib.auth.forms import (
     ReadOnlyPasswordHashField, ReadOnlyPasswordHashWidget,
-    PasswordResetForm as OldPasswordResetForm
+    PasswordResetForm as OldPasswordResetForm,
+    UserChangeForm as DjangoUserChangeForm,
+    AuthenticationForm as DjangoAuthenticationForm,
 )
-from django.contrib.auth import get_user_model
-from django.contrib.auth.hashers import identify_hasher
-from django.utils.translation import ugettext_lazy as _, ugettext
+from django.contrib.auth import get_user_model, password_validation
+from django.contrib.auth.hashers import identify_hasher, UNUSABLE_PASSWORD_PREFIX
+from django.utils.translation import gettext_lazy as _, gettext
 from django.utils.html import format_html
 
 User = get_user_model()
 
 
 def is_password_usable(pw):
-    # like Django's is_password_usable, but only checks for unusable
-    # passwords, not invalidly encoded passwords too.
-    try:
-        # 1.5
-        from django.contrib.auth.hashers import UNUSABLE_PASSWORD
-        return pw != UNUSABLE_PASSWORD
-    except ImportError:
-        # 1.6
-        from django.contrib.auth.hashers import UNUSABLE_PASSWORD_PREFIX
-        return not pw.startswith(UNUSABLE_PASSWORD_PREFIX)
+    """Decide whether a password is usable only by the unusable password prefix.
+
+    We can't use django.contrib.auth.hashers.is_password_usable either, because
+    it not only checks against the unusable password, but checks for a valid
+    hasher too. We need different error messages in those cases.
+    """
+
+    return not pw.startswith(UNUSABLE_PASSWORD_PREFIX)
 
 
 class BetterReadOnlyPasswordHashWidget(ReadOnlyPasswordHashWidget):
     """
     A ReadOnlyPasswordHashWidget that has a less intimidating output.
     """
-    def render(self, name, value, attrs):
-        from django.forms.util import flatatt
-        final_attrs = flatatt(self.build_attrs(attrs))
+    def render(self, name, value, attrs=None, renderer=None):
+        final_attrs = flatatt(self.build_attrs(attrs or {}))
 
         if not value or not is_password_usable(value):
-            summary = ugettext("No password set.")
+            summary = gettext("No password set.")
         else:
             try:
                 identify_hasher(value)
             except ValueError:
-                summary = ugettext("Invalid password format or unknown"
+                summary = gettext("Invalid password format or unknown"
                                    " hashing algorithm.")
             else:
-                summary = ugettext('*************')
+                summary = gettext('*************')
 
         return format_html('<div{attrs}><strong>{summary}</strong></div>',
                            attrs=final_attrs, summary=summary)
 
 
+class UserChangeForm(DjangoUserChangeForm):
+    def __init__(self, *args, **kwargs):
+        super(UserChangeForm, self).__init__(*args, **kwargs)
+        password = self.fields.get('password')
+        if password:
+            password.widget = BetterReadOnlyPasswordHashWidget()
+
+
 class UserCreationForm(forms.ModelForm):
     """
     A form for creating new users. Includes all the required
@@ -94,6 +102,17 @@ class UserCreationForm(forms.ModelForm):
             raise forms.ValidationError(self.error_messages['password_mismatch'])
         return password2
 
+    def _post_clean(self):
+        super(UserCreationForm, self)._post_clean()
+        # Validate the password after self.instance is updated with form data
+        # by super().
+        password = self.cleaned_data.get('password2')
+        if password:
+            try:
+                password_validation.validate_password(password, self.instance)
+            except forms.ValidationError as error:
+                self.add_error('password2', error)
+
     def save(self, commit=True):
         # Save the provided password in hashed format
         user = super(UserCreationForm, self).save(commit=False)
@@ -103,41 +122,28 @@ class UserCreationForm(forms.ModelForm):
         return user
 
 
-class UserChangeForm(forms.ModelForm):
+class CaseInsensitiveUsernameFieldCreationForm(UserCreationForm):
     """
-    A form for updating users. Includes all the fields on
-    the user, but replaces the password field with admin's
-    password hash display field.
+    This form is the same as UserCreationForm, except that usernames are lowercased before they
+    are saved. This is to disallow the existence of email address usernames which differ only in
+    case.
     """
-    password = ReadOnlyPasswordHashField(label=_("Password"),
-        widget=BetterReadOnlyPasswordHashWidget)
+    def clean_USERNAME_FIELD(self):
+        username = self.cleaned_data.get(User.USERNAME_FIELD)
+        if username:
+            username = username.lower()
 
-    class Meta:
-        model = User
-        if DJANGO_VERSION >= (1, 6):
-            fields = '__all__'
+        return username
 
-    def __init__(self, *args, **kwargs):
-        super(UserChangeForm, self).__init__(*args, **kwargs)
-        f = self.fields.get('user_permissions', None)
-        if f is not None:
-            f.queryset = f.queryset.select_related('content_type')
+# set the correct clean method on the class so that child classes can override and call super()
+setattr(
+    CaseInsensitiveUsernameFieldCreationForm,
+    'clean_' + User.USERNAME_FIELD,
+    CaseInsensitiveUsernameFieldCreationForm.clean_USERNAME_FIELD
+)
 
-    def clean_password(self):
-        # Regardless of what the user provides, return the initial value.
-        # This is done here, rather than on the field, because the
-        # field does not have access to the initial value
-        return self.initial["password"]
-
-
-class AdminUserChangeForm(UserChangeForm):
-    def __init__(self, *args, **kwargs):
-        super(AdminUserChangeForm, self).__init__(*args, **kwargs)
-        if not self.fields['password'].help_text:
-            self.fields['password'].help_text = _(
-                "Raw passwords are not stored, so there is no way to see this"
-                " user's password, but you can change the password using"
-                " <a href=\"password/\">this form</a>.")
+# alias for the old name for backwards-compatability
+CaseInsensitiveEmailUserCreationForm = CaseInsensitiveUsernameFieldCreationForm
 
 
 class FriendlyPasswordResetForm(OldPasswordResetForm):
@@ -147,19 +153,27 @@ class FriendlyPasswordResetForm(OldPasswordResetForm):
                                   "sure you've registered?")
 
     def clean_email(self):
-        super_clean_email = getattr(
-            super(FriendlyPasswordResetForm, self), 'clean_email', None)
-        if callable(super_clean_email):  # Django == 1.5
-            # Django 1.5 sets self.user_cache
-            return super_clean_email()
+        """Return an error message if the email address being reset is unknown.
+
+        This is to revert https://code.djangoproject.com/ticket/19758
+        The bug #19758 tries not to leak emails through password reset because
+        only usernames are unique in Django's default user model.
+
+        django-authtools leaks email addresses through the registration form.
+        In the case of django-authtools not warning the user doesn't add any
+        security, and worsen user experience.
+        """
 
-        # Simulate Django 1.5 behavior in Django >= 1.6.
-        # This is not as efficient as in Django 1.5, since clean_email() and
-        # save() will be running the same query twice.
-        # Whereas Django 1.5 just caches it.
         email = self.cleaned_data['email']
-        qs = User._default_manager.filter(is_active=True, email__iexact=email)
-        results = [user for user in qs if user.has_usable_password()]
+        results = list(self.get_users(email))
+
         if not results:
             raise forms.ValidationError(self.error_messages['unknown'])
         return email
+
+
+class AuthenticationForm(DjangoAuthenticationForm):
+    def __init__(self, request=None, *args, **kwargs):
+        super(AuthenticationForm, self).__init__(request, *args, **kwargs)
+        username_field = User._meta.get_field(User.USERNAME_FIELD)
+        self.fields['username'].widget = username_field.formfield().widget
